﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection.Emit;

namespace EmitExamples
{
    class Animal
    {
        public virtual void Speak()
        {
            Console.WriteLine("Animal.Speak");
        }
    }

    class Cat : Animal
    {
        public override void Speak()
        {
            Console.WriteLine("Cat.Speak");
        }
    }

    class Dog : Animal
    {
        public override void Speak()
        {
            Console.WriteLine("Dog.Speak");
        }
    }

    class Program
    {
        private delegate void SpeakDelegate(Animal animal);

        static void Main(string[] args)
        {
            //定义动态方法，没有返回值，传入参数为Animal，所在的模块选择为Program类所在的模块
            DynamicMethod dynamicSpeakWithCall = new DynamicMethod("DynamicSpeakWithCall", null, new Type[] { typeof(Animal) }, typeof(Program).Module);

            ILGenerator callIL = dynamicSpeakWithCall.GetILGenerator();

            //加载参数0 即 Animal类或其派生类的对象
            callIL.Emit(OpCodes.Ldarg_0);
            //通过Call指令调用Speak方法
            callIL.Emit(OpCodes.Call, typeof(Animal).GetMethod("Speak"));
            callIL.Emit(OpCodes.Ret);

            Console.WriteLine("SpeakWithCall:");
            SpeakDelegate SpeakWithCall = (SpeakDelegate)dynamicSpeakWithCall.CreateDelegate(typeof(SpeakDelegate));
            SpeakWithCall(new Animal());
            SpeakWithCall(new Cat());
            SpeakWithCall(new Dog());

            //定义动态方法，没有返回值，传入参数为Animal，所在的模块选择为Program类所在的模块
            DynamicMethod dynamicSpeakWithCallvirt = new DynamicMethod("DynamicSpeakWithCallvirt", null, new Type[] { typeof(Animal) }, typeof(Program).Module);

            ILGenerator callvirtIL = dynamicSpeakWithCallvirt.GetILGenerator();

            //加载参数0 即 Animal类或其派生类的对象
            callvirtIL.Emit(OpCodes.Ldarg_0);
            //通过Callvirt指令调用Speak方法
            callvirtIL.Emit(OpCodes.Callvirt, typeof(Animal).GetMethod("Speak"));
            callvirtIL.Emit(OpCodes.Ret);

            Console.WriteLine("SpeakWithCallvirt:");
            SpeakDelegate SpeakWithCallvirt = (SpeakDelegate)dynamicSpeakWithCallvirt.CreateDelegate(typeof(SpeakDelegate));
            SpeakWithCallvirt(new Animal());
            SpeakWithCallvirt(new Cat());
            SpeakWithCallvirt(new Dog());
        }
    }
}
